\ CONTENT : contains the factorization object
\ LANGUAGE : tForth
\ PROJECT : hcc primes
\ CATEGORY : EXAMPLE
\ AUTOR :  AH
\ LAST CHANGE : 911118 AH could only find first composite factor
\ LAST CHANGE : 911118 AH better not use SS as a scratch here
\ LAST CHANGE : 911115 AH created

\ Object factorization :
\   list of { times, composite, len, content }
\    if times is 0 , this marks the end of the list
\    times indicates how many times the factor with a `content' of 'len'
\     bytes is present in the number represented by this factorization.
\   composite is a flag, indicating that the factor is composite still
\
\ Messages :
\     %clear   set the f~ to empty
\  V  %to      put a Vnumber into the f~
\     %print   print the f~
\  i  %prime   divide out the prime from f~
\  V  %factor  divide out a factor maybe not prime
\  V  %todo    returns a number that still is to be factorized
\                before this factorization is completed
\     %?factorized returns a boolean indicating whether factorized

\ Estimate space needed for a factorization
\ about the same as in the direct representation of the number
MaxV 1+ CELLS =: #FAC

31 =:  %factor%      \ divide out a factor maybe not prime
32 =:  %todo%        \ returns a number that still is to be factorized
33 =:  %?factorized% \ returns a boolean indicating whether factorized

: %factor        %factor%      %VAR !  ;
: %todo          %todo%        %VAR !  ;
: %?factorized   %?factorized% %VAR !  ;


        REVISION -f " Factorizations      Version 1.00 "


WARNING @  WARNING OFF
  CREATE BUFFER #FAC ALLOT      \ Auxiliary, in fact a Factorization
WARNING !

Vnumber VFscratch1              \ Need scratch, separate from everything
Vnumber VFscratch2              \ Need scratch, separate from everything

\ BACKSPACE erases the last character printed on the screen
: BACKSPACE ^H EMIT BL EMIT ;

\ Get VFscratch1 from addres A1 with auto increment, leave the incremented address
\ Format of source like in a factorization object
: VFscratch1@+             ( A1 -- A2 )
        0 VFscratch1 !
        C@+ 0 ?DO              \ Third byte length
                VFscratch1 8 V<<           \ Accumulate bytes in VFscratch1
                C@+ VFscratch1 SWAP VS+
        LOOP
;

: C!+     ( address content -- incremented address )
   OVER C!
   1+
;

\ Put VFscratch1 to addres A1 with auto increment, leave the incremented address
\ Format of destination like above
: VFscratch1!+             ( A1 -- A2 )
        LOCAL Address
        VFscratch1 Vcount-bits 8 /MOD LOCAL Length
        ( Remainder) IF Length 1+ TO Length THEN

        Length Address C!

        1 Length DO
           VFscratch1 256 VS/MOD
           Address I + C!
        -1 +LOOP

        Address 1+ Length +
;

\ Put N to addres A1 with auto increment, leave the incremented address
\ Format of destination like above
: N!+             ( A1 N -- A2 )
        LOCAL N
        LOCAL Address
        N Scount-bits 8 /MOD LOCAL Length
        ( Remainder) IF Length 1+ TO Length THEN

        Length Address C!

        1 Length DO
           N U>D 256 UM/MOD TO N
           Address I + C!               \ Put remainder next digit
        -1 +LOOP

        Address 1+ Length +
;

\ Executes the print message
\ Print factorization with a mathematical format and in decimal :
\ Example : c12*3^6*19*c3.333.456
\ c's announce a composite factor
: PrintFac            ( F -- )
        0 LOCAL Power

   BASE @ >R DECIMAL

   BEGIN
        C@+ TO Power           \ First byte Power
        Power WHILE
        C@+ IF &c EMIT THEN    \ Second byte: composite flag
        VFscratch1@+ VFscratch1 .VNUM
        Power 1 <> IF
            &^ EMIT Power .
            ^H EMIT             \ Rid of space that always accompanies .
        ENDIF
        &* EMIT
   REPEAT
   BACKSPACE            \ Get rid of last star
   DROP       \ Address
   R> BASE ! DECIMAL
;

\ Execute the TO message
\ Put the Vnumber in a factorization
: VnumToFac   ( V,F -- )
     SWAP LOCAL V

     V @ 0= ABORT" Forbidden to factorize 0 "

     1 C!+              \ To the power 1
     V ?Vprime 0= C!+    \ Composite indicator
     V VFscratch1 VVmove
     VFscratch1!+
     0 C!+              \ Closing end
     DROP               \ address
;


\ Execute the TODO message
\ Fetches the first composite factor in the factorization
: VnumFromFac   ( F,V -- )
    SWAP LOCAL V

    BEGIN
        C@+ WHILE
        C@+ IF                  \ Is the factor composite?
           VFscratch1@+
           VFscratch1 V VVmove
           DROP                 \ The address
           EXIT
        ELSE
           VFscratch1@+         \ Skip
        ENDIF
   REPEAT
   TRUE ABORT" Fetching a to do from a fully factored number"
;


\ TrySmallDivider tries the prime divider p on all the composite
\ factors of F
: TrySmallDivider      ( p, F -- )
        LOCAL Factorization     \ Move from
        LOCAL Prime             \ The test prime
BUFFER  LOCAL Out               \ Move to
0       LOCAL ?comp             \ the composite indicator
0       LOCAL Power             \ Power of some number

   Factorization          \ This is an incrementing fetch underground

   BEGIN
        C@+ TO Power           \ First byte Power
        Power WHILE
        C@+ TO ?comp           \ Second byte: composite flag
        VFscratch1@+
        0
        BEGIN
\            CR VFscratch1 .VNUM Prime . ^
            VFscratch1 Prime VSMOD 0= WHILE
            VFscratch1 Prime VS/MOD DROP \ Remove Prime from VFscratch1
            1+
        REPEAT
        DUP IF                   \ Is there a factor Prime
            Out SWAP Power * C!+ \ Power
            0 C!+                \ Composite indicator
            Prime N!+            \ Number
            TO Out
             VFscratch1 ?Vprime 0= TO ?comp \ Recalculate compositeness
        ELSE
            DROP
        THEN

        VFscratch1 1 VS= 0= IF        \ In seldom cases a 1 may remain
            Out Power C!+
            ?comp C!+
            VFscratch1!+
            TO Out
        THEN
   REPEAT
   DROP
   0 Out C!

   \ Move back from scratch area
   BUFFER Factorization #FAC CMOVE

;


\ TryBigDivider tries the divider V on all the composite
\ factors of F
: TryBigDivider      ( V, F -- )
        LOCAL Factorization     \ Move from
        LOCAL V                 \ The factor, a VNUM
BUFFER  LOCAL Out               \ Move to
0       LOCAL ?comp             \ the composite indicator
0       LOCAL Power             \ Power of some number

   Factorization          \ This is an incrementing fetch underground

   BEGIN
        C@+ TO Power           \ First byte Power
        Power WHILE
        C@+ TO ?comp           \ Second byte: composite flag
        VFscratch1@+

        ?comp IF                \ Need only be done if Vfac is composite
            0                       \ Count the factors
            BEGIN
                VFscratch1 V VV/MOD RR @ 0= WHILE
                QQ VFscratch1 VVmove       \ Remove divider from VFscratch1
                1+
            REPEAT
            DUP IF                   \ Did the factor indeed occur
                Out SWAP Power * C!+ \ Power
                V ?Vprime 0= C!+     \ Composite indicator
                VFscratch1 VFscratch2 VVmove       \ Temporarily save VFscratch1
                V VFscratch1 VVmove       \ Store the factor in it
                VFscratch1!+              \ Number
                VFscratch2 VFscratch1 VVmove      \ Restore VFscratch1
                TO Out
            ELSE
                DROP
            THEN
        THEN

        VFscratch1 1 VS= 0= IF        \ In seldom cases a 1 may remain
            Out Power C!+
            VFscratch1 ?Vprime 0= C!+
            VFscratch1!+
            TO Out
        THEN
   REPEAT
   DROP
   0 Out C!

   \ Move back from scratch area
   BUFFER Factorization #FAC CMOVE

;


\ Test whether the factorization has been completed
: ?Factorized   ( F -- flag )
    TRUE LOCAL Factorized               \ Innocent until proven guilty
    BEGIN
        C@+ WHILE
        C@+ 0= Factorized AND TO Factorized
        C@+             \ Skip length
        +               \ Skip length bytes content
   REPEAT
   DROP                 \ Skip address
   Factorized
;


\ Make the factorization object
: CreateFactorization CREATE                          \ <> C~ #<name># \> <>
                   #FAC ALLOT
                DOES>                           \ <??> <pin> -- <?>
                   %VAR @  0 %VAR !
                   CASE
                      %clear%       OF 0 SWAP !               ENDOF
                      %factor%      OF TryBigDivider          ENDOF
                      %prime%       OF TrySmallDivider        ENDOF
                      %addressof%   OF  ( Leave address)      ENDOF
                      %to%          OF VnumToFac              ENDOF
                      %todo%        OF VnumFromFac            ENDOF
                      %print%       OF PrintFac               ENDOF
                      %?factorized% OF ?Factorized           ENDOF
                      DUP DEC. TRUE ABORT" PBASE: unknown message"
                   ENDCASE
;
