\  LANGUAGE    : ANSI FORTH with TFORTH peculiarities
\  PROJECT     : Factoring
\  CATEGORY    : main source file
\  AUTHOR      : Marcel Hendrix, 910921
\  LAST CHANGE : 931127 AH <REPACK removed
\  LAST CHANGE : 931126 AH pop, removed
\  LAST CHANGE : 930410 AH VRoot added
\  LAST CHANGE : 930302 AH SHIFT updates (changes to DPANS)
\  LAST CHANGE : ah 911118 Scount-bits okay for input 0
\  LAST CHANGE : ah 911115 configuration moved to CONFIG.FRT
\  LAST CHANGE : ah 911112 .VNUM added
\  LAST CHANGE : ah 911110 TheRoot added
\  LAST CHANGE : ah 911109 Prime base separated
\  LAST CHANGE : ah 911108 n- bug in VV*
\  LAST CHANGE : ah 911107 Anti bugging, Sqrt, normalization
\  LAST CHANGE : ah 911030 PP QQ RR SS introduced
\  LAST CHANGE : ah 911030 testing VV/MOD works!
\  LAST CHANGE : ah 911028 a dozen ERRORS IN vv/mod MUL&SUB
\
\  LAST CHANGE :           October 25, 1991, Albert van der Horst
\                 bug n VV* caused by order change
\  LAST CHANGE : 00:50:18  October 12, 1991, Marcel Hendrix
\

\ This version stands overnight testing of ?RABBIN


        REVISION -factors "--- Factors module 1&2  Version 3    ---"


                                \ Utility \

S" TFORTH"  ENVIRONMENT? [IF] DROP

\ Subtract two single numbers with borrow.
\ Works only under the transputer Forth
        CODE UM-                \ <borrow> <u1> <u2> -- <u> <borrow>
           >pr0,
            dpop, >pr1,
            dpop, dpop, rev,      \ get borrow in CR:
            pr1>,
            ldiff,              \ BR: <borrow>  AR: <sum>
            dpush, dpush,
           pr0>, gcall,
        END-CODE


[ELSE]

\ Alternative non-code definition :
\ WARNING : NOT TESTED
\ u1-u2-oldborrow is mathematisch gelijk aan unew- borrow*<base>
: UM-                           \ <oldborrow> <u1> <u2> -- <unew>  <newborrow>
   >R >R U>D R> U>D R> U>D      \ Everything double
   D- 2SWAP D-
   NEGATE                       \ Only the least significant byte of D
;

[THEN]

                           \ MODULE 1 ** DATA \





DOC Module 1
(*
     Sizes:
     MaxV represents the maximum length in cells of the number
     created by Vnumber.
     MaxB is the length in cells of a bit array.
     MaxPrime is the limit of the numbers used in sieving.

     data structures:

     Vnumber  ( "name" -- )
     creates a data structure that represents a variable length number,
     of at most MaxV cells.
     First cell contains the actual length.
     The least significant digits are leftmost.
     Example : $0000 0002  $FFFF FFFF $0000 0001 XX XX
     represents 2^33 -1 with XX don't care.
     If you do not have an ANSI Forth, define
     : CELLS 4 * ; and : CELL+ 4 + ; for a 32 bit forth
     : CELLS 2 * ; and : CELL+ 2 + ; for a 16 bit forth
     Use those whenever manipulating addresses.

     The most important VNumber is :
            TheNumber
     This one is going to be factored.
     There are two scratch VNum's : VScratch1 and VScratch2

     PrimeBase
     This is a TO-object.
     It is an array of structs, length to be determined
     In the following: Pin means prime-index.

     Pin %prime PrimeBase
        delivers a prime or prime power (p^n) or 0 (out of range)

     Prime %to PrimeBase
        puts the prime in the next free place of the database

     Pin %2log PrimeBase
        delivers the log of that prime or prime power (p^n)

     Start Pin %tostart PrimeBase
           Pin %fromstart PrimeBase
        allow to store and retrieve for each Pin an auxiliary number
        in the PrimeBase called Start.
*)
ENDDOC


\ ************* PART 1 Lengths ***********************

1 CELLS 8 * CONSTANT BITS/CELL

\ For each possible prime there is some adjustable number of powers
MaxB BITS/CELL * #ProportPrPo * =: MaxPrime  \ Maximum # entries in prime base.

VALUE MaxBase       \ The largest prime or power we allow in the factor base

: Vnumber       CREATE  0 , MaxV CELLS ALLOT    \ <> Vnumber #<name># \> <>
                DOES>   ;                       \ <> -- <addr>

        Vnumber TheNumber
        Vnumber TheRoot
        Vnumber TheFactor
        TheNumber =: TT         \ Alias
        Vnumber PP              \ Product
        Vnumber QQ              \ Quotient
        Vnumber RR              \ Remainder
        Vnumber SS              \ Square root, Scratch & GCD


                        \ MODULE 2 ** BASIC MANIPULATIONS \




DOC
(*
 Module 2
     In the following vocabulary "V" means the address of a Vnum,
     "u" a unsigned single precision Forth word.

     Stack diagram ( V u -- )
     VS-          subtract u from V (in place)
     VS+          add u to V (in place)
     VS*          multiply V with u (in place)

     Stack diagram ( V u -- flag )
     VS=
     Stack diagram ( V u1 -- u2 )
     VS/MOD       as /MOD , quotient in V, remainder in u2
     VSMOD        as MOD, remainder in u2, V untouched

     Stack diagram ( V -- n )
     Vcount-bits  count how many siginificant bits the number V has

     Stack diagram ( V -- )
     VSqrt        Calculate the largest number whose square is smaller
                  than V, put it in SS
     1V!          Put the constant 1 in V

     Stack diagram ( V1 V2 -- )
     VVmove       copy V1 to V2
     VV-          subtract V2 from V1 (in place)
     VV+          add V2 to V1 (in place)
     VV*          multiply V1 and V2, result to PP
     VV/MOD       as /MOD , quotient in QQ, remainder in RR
     VGCD         greatest common divisor of V1 and V2, result in SS

     Most are straightforward.

     VV* benefits from the formula : (Knuth part 3 pg. 278)
     ( 2^n*u1+u0)*(2^n*v1+v0) =
     ( 2^2n+2^n)*u1*v1 + 2^n*(u1-u0)*(v1-v0) + (2^n+1)*u0*v0.
     (saves 1 out of 4 multiplications)

     VV/MOD can be implemented as a normal tail division, i.e. find
     a single precision multiplier D for V2 such that subtracting D*V2
     from the front of V1 makes it first cell zero, so D is the first
     cell of the quotient.
     Contrary to the expectation, this is about the best you can do.

     VGCD can be implemented using VV/MOD but an add and shift algorithm
     is probably faster.
*)
ENDDOC

\ ************* PART 1 Auxiliary definitions ***********************

\ Store 1 in the Vnumber
: 1V!   ( Vnum -- )
    1 OVER !   CELL+
    1 SWAP !
;

\ Returns the number of bits in the number U
: Scount-bits     ( U -- I )
  DUP 0= IF DROP 0 EXIT THEN
  0
  BEGIN 1+ SWAP 1 RSHIFT DUP WHILE SWAP REPEAT
  DROP
;

\ Returns the number of bits in the Vnumber V
: Vcount-bits     ( V -- I )
  DUP @ 1- BITS/CELL *
  SWAP DUP @ CELLS + @          \ Get last cell
  Scount-bits                   \
  +
;

\ ************* PART 2 The elementary arithmetic functions ***********************

\ Add u to V in place. No bounds check; if carry -> 1 digit longer.
\ We assume the least significant digit is at the lowest memory address.

: VS+               LOCAL carry                 \ <V> <u> -- <>
                DUP LOCAL V
                @+ 0 ?DO
                        @+  carry UM+ TO carry
                        OVER CELL- !
                    LOOP
                carry IF  carry SWAP !  1 V +!  EXIT
                   ENDIF
                DROP ;


\ Subtract u from V in place. The number may become 1 digit shorter.

: (VS-)             LOCAL u                     \ <V> <u> -- <borrow>
                  0 LOCAL borrow
                DUP LOCAL V
                @+ 0 ?DO
                         @+ borrow SWAP u UM- TO borrow
                         OVER CELL- !
                         CLEAR u
                    LOOP
                CELL- @ 0= IF -1 V +!
                        ENDIF
                borrow ;


\ Warning an unnormalized number may result from VS-
: VS-           (VS-) IF CR ." VS- : Overflow " \ <V> <u> -- <borrow>
                  ENDIF
;



\ Test if the Vnumber is 0.

: VS0=          @ 0= ;                          \ <V> -- <bool>


\ multiply V with u in place. The length of the result can be zero, equal, or
\ 1 digit larger.

: VS*           DUP 0= IF SWAP ! EXIT ENDIF     \ <V> <u> -- <>
                    LOCAL u
                  0 LOCAL carry
                DUP LOCAL V
                @+ 0 ?DO
                        @+ u UM*
                        carry U>D D+ TO carry
                        OVER CELL- !
                    LOOP
                carry ?DUP IF  SWAP !  1 V +!  EXIT
                        ENDIF
                DROP ;


\ copy V1 to V2 ( they may overlap)

: VVmove        OVER @ 1+ CELLS  CMOVE ;        \ <V1> <V2> -- <>

\ Compare the two V-number
: VV<=           ( V1,V2 -- flag )
        LOCAL V2
        LOCAL V1

        V1 @ V2 @ <> IF
           V1 @ V2 @ <   \ Number of digits preveal over content
        ELSE
           TRUE         \ If we do not leave the loop : equal
           1 V1 @  DO   \ Compare digits, starting with most significant
              V1 I CELLS + @
              V2 I CELLS + @ 2DUP <> IF
                 < NIP LEAVE
              ELSE
                  2DROP
              ENDIF
           -1 +LOOP
        ENDIF
;

\ Try to remove zeroes in the most significant digit positions.

: COMPACT       >S S @+ CELLS +                 \ <V> -- <>
                S @ 0 ?DO
                          CELL- DUP @ IF LEAVE THEN
                          -1 S +!
                     LOOP
                DROP -S ;


\ as MOD, remainder in u2, V untouched
: VSMOD         OVER @ 0=  OVER 0= OR           \ <V> <u1> -- <u2>
                IF 2DROP 0 EXIT ENDIF           \ already zero or div by 0?
                0 LOCAL carry
                  LOCAL u1
                  LOCAL V
                V @ CELLS  V +                  \ '(latest digit)
                V @ FOR
                      AFT
                        @- carry  u1 UM/MOD
                        SWAP TO carry
                        DROP
                     THEN
                   NEXT
                DROP
                carry ;

\ as /MOD , quotient in V, remainder in u2. 1 <= New length of V <= old len.

: VS/MOD        OVER @ 0=  OVER 0= OR           \ <V> <u1> -- <u2>
                IF 2DROP 0 EXIT ENDIF           \ already zero or div by 0?
                0 LOCAL carry
                  LOCAL u1
                  LOCAL V
                V @ CELLS  V +                  \ '(latest digit)
                V @ FOR
                      AFT
                        @- carry  u1 UM/MOD
                        SWAP TO carry
                        OVER CELL+ !
                     THEN
                   NEXT
                DROP
                V COMPACT carry ;


\ Is V1 equal to V2 ?

: VV=           @+ >S SWAP @+ S>                \ <V1> <V2> -- <bool>
                2DUP <> IF  2DROP 2DROP  FALSE EXIT  THEN
                DROP CELLS TUCK COMPARE 0= ;


\ Give a Vnum at least u digits by extending it with zeros.
\ Count is adjusted!

: Vextend       >S DUP @+ >S                    \ <V> <u> -- <>
                T S > IF S CELLS +
                         T S> - CELLS ERASE     \ put in 0's
                         S> SWAP !              \ update count
                    ELSE 2DROP -S -S            \ no action
                   ENDIF ;


\ Add V2 to V1 in place. V1 may grow to max(V1,V2)+1.

: VV+              0 LOCAL carry                \ <V1> <V2> -- <>
                     LOCAL V2
                     LOCAL V1
                V2 @ LOCAL #V2
                V1 V2 @  Vextend                \ give them the same length
                V2 V1 @  Vextend
                V1 @+ DROP LOCAL p1
                V2 @+ SWAP LOCAL p2
                  0 ?DO
                        p1 @+ SWAP TO p1
                        p2 @+ SWAP TO p2
                        UM+  carry U>D D+  TO carry
                        p1 CELL- !
                   LOOP
                #V2 V2 !                        \ reset length of V2
                carry ?DUP IF  p1 !  1 V1 +!  EXIT
                        ENDIF ;


\ Subtract V2 from V1 in place. V2 is not modified. 0 <= Length V1 <= V1

: (VV-)            0 LOCAL borrow               \ <V1> <V2> -- <borrow>
                     LOCAL V2
                     LOCAL V1
                V2 @ LOCAL #V2
                V1 V2 @  Vextend                \ give them the same length
                V2 V1 @  Vextend
                V1 @+ DROP LOCAL p1
                V2 @+ SWAP LOCAL p2
                  0 ?DO
                        borrow
                        p1 @+ SWAP TO p1
                        p2 @+ SWAP TO p2
                        UM- TO borrow
                        p1 CELL- !
                   LOOP
                #V2 V2 !                        \ reset length of V2
                V1 COMPACT
                borrow ;


: VV-           (VV-) IF CR ." VV- : overflow " \ <V1> <V2> -- <>
                   ENDIF ;



\ **************** The more difficult arithmetics *****************
\ multiply V1 and V2, result to PP.
\ See:     `The Art of Computer Programming' Second Edition,
\ Volume 2, Seminumerical Algorithms, D. E. Knuth, Algorithm M, pp 253-254.


: VV*           PP LOCAL w               \ <Un> <Vm> -- <>
                          LOCAL v
                      v @ LOCAL m
                          LOCAL u
                      u @ LOCAL n
                        0 LOCAL carry
                m 0=  n 0= OR IF  0 w ! EXIT
                           ENDIF
                m n +  w !
                w CELL+  n CELLS ERASE          \ M1
                m 1+ 1 DO  I v []CELL  @
                           0= IF 0 I n + w []CELL ! \ M2
                            ELSE CLEAR carry    \ M3
                                 n 1+ 1 DO      \ M4
                                           I u []CELL @
                                           J v []CELL @  UM*
                                           I J + 1- w []CELL >S  S @  U>D D+
                                           carry U>D D+ TO carry
                                           S> !
                                      LOOP      \ M5
                                 carry  I n + w []CELL !
\   AH 91 10 24                           ^^^
                           ENDIF
                     LOOP                       \ M6
                w COMPACT ;


\ Shift a Vnum up over u bit positions. ( u < bitsperword)

: V<<             0 LOCAL carry                 \ <V> <u> -- <>
                    LOCAL places
                DUP LOCAL V
                @+
                0 ?DO
                      @+ U>D  places DLSHIFT  carry U>D D+  TO carry
                      OVER CELL- !
                 LOOP
                carry IF carry SWAP !  1 V +! EXIT
                   ENDIF
                DROP ;


\ Shift a Vnum down over u bit positions. ( u < bitsperword)

: V>>             0 LOCAL carry                 \ <V> <u> -- <>
                    LOCAL places
                DUP LOCAL V
                @+ >S S 1- CELLS + S>
                0 ?DO
                      @- 0 SWAP  places DRSHIFT  0 carry D+  SWAP TO carry
                      OVER CELL+ !
                 LOOP
                ( v) COMPACT ;


\ Convert Vnum to exponent, mantissa format
\ i.e.  V{in} = v{out}/2^exponent
\ and the m.s. bit ( of the m.s. cell) of v{out} is 1
: VNORMALIZE      0 LOCAL exp                   \ <V> -- <exponent>
                DUP LOCAL V
                @+ 1- CELLS + @                 \ highest word
                DUP 0= IF EXIT        ENDIF     \ impossible!
                DUP 0< IF DROP 0 EXIT ENDIF     \ already normalized
                BEGIN
                   1 LSHIFT  1 +TO exp
                   DUP 0<
                UNTIL
                ( hi word) DROP
                V exp V<<  exp ;


\ Convert exponent, mantissa format to Vnum

: VUNNORMALIZE  V>> ;                           \ <V> <exp> -- <>


\ Like /MOD . Quotient in QQ, remainder in RR.
\ Knuth's Algorithm D.

\ U should have at least two digits!

\ May be used for debugging
: ^^ CR ." dst " 20 DUMP ." src " 20 DUMP ;

\ Auxiliary definition for VV/MOD
\ Sorely needs an exact description
: MUL&SUB           LOCAL qhat          \ <dst> <src> <u> <qhat> -- <borrow>
                ROT LOCAL 'dest
                  0 LOCAL carry
                  0 LOCAL borrow
                0 ?DO
                     @+ qhat UM* carry U>D D+ TO carry
                     borrow 'dest @ ROT UM- TO borrow  'dest !
                     1 CELLS +TO 'dest
                 LOOP
                DROP
                borrow 'dest @ carry UM- SWAP 'dest !
                ( Leaves borrow ) ;

\ Auxiliary definition for VV/MOD
\ Sorely needs an exact description
: ADDBACK       ROT LOCAL 'dest                       \ <dst> <src> <u> -- <>
                  0 LOCAL carry
                0 ?DO
                     @+ 'dest @ UM+  carry U>D D+ TO carry
                     'dest !
                     1 CELLS +TO 'dest
                 LOOP
                 1 carry <> IF ." No carry in ADDBACK " THEN
                 'dest @ -1 <> IF ." No second carry in addback" THEN
                 carry 'dest +!
                 [ DEBUG ] [IF] 7 EMIT [THEN]
                DROP ;

 VALUE U1 VALUE U2                 \ To be restored to locals
 VALUE V1 VALUE V2
 VALUE qhat
: (vv/mod)
          LOCAL V                               \ <u> <v> -- <>
          LOCAL U
            U    RR VVmove
                    0        TO qhat
                RR @ V @ -    1+ LOCAL m+1  m+1 QQ !
                V VNORMALIZE LOCAL exp
                V @ CELLS V +  @- TO V1
                                  @  TO V2           \ highest digits of V
                RR  RR @ 1+  Vextend                      \ D1.
                RR exp V<<
                RR @ CELLS RR +                           \ D2.
                m+1 0 ?DO
                        DUP @ TO U1
                        DUP 1 CELLS - @ TO U2
                        DUP @ V1 = IF                     \ D3.
                           -1 TO qhat
                        ELSE
                            DUP CELL- 2@ SWAP V1 UM/MOD  TO qhat DROP
                        ENDIF ( 'uj )
                        BEGIN
                           DUP CELL- 2@ SWAP  qhat V1 UM* D-   \
                           >R        \ Save m.s. cell, if non-zero no
                                      \ decrementing of qhat is needed
                           OVER 2 CELLS - @  SWAP  \ l.s. result of D- and Uj+2 from double together
                               V2 qhat UM*   DU<
                               R> 0= AND           \ And also m.s. cell 0
                        WHILE  -1 +TO qhat
                        REPEAT
                        [ DEBUG ] [IF] CR qhat . [THEN]
                        DUP  V @ CELLS -  V CELL+  V @
                           qhat MUL&SUB               \ D4.
                        IF -1 +TO qhat               \ D6.
                              DUP V @ CELLS -  V CELL+
                              V @  ADDBACK
                        ENDIF
                        qhat m+1 I - QQ []CELL  !
[ DEBUG ] [IF]
                        DUP @ IF
                          ." Next digit in RR is not zeroed at index " I .

                          DECIMAL
                          CR ." U " U .VNUM
                          CR ." V " V .VNUM
                          RR 60 DUMP ABORT
                        THEN
[THEN]
                        CELL-                           \ D7.
                 LOOP
                DROP
                RR exp VUNNORMALIZE                      \ D8.
                V exp VUNNORMALIZE
                QQ COMPACT   RR COMPACT   \ Should not be necessary....
          ;



: VV/MOD        DUP @ 0=                        \ <V1> <V2> -- <>
                   IF 2DROP
                      0 QQ !  0 RR !
                      EXIT
                ENDIF
                DUP @ 1 =
                   IF SWAP QQ VVmove
                      QQ SWAP CELL+ @ VS/MOD
                      ?DUP IF 1 RR !  RR CELL+ !
                         ELSE 0 RR !  ( special case : 0 ...)
                        ENDIF
                      EXIT
                ENDIF
                OVER @  OVER @ < IF DROP 0 QQ !  RR VVmove
                                    EXIT
                              ENDIF
                (vv/mod) ;


\ ************* PART 3 The compound arithmetic functions ***********************

\ Greatest common divisor of U and V, result in SS
\   A1. v = 0 => u is answer
\   A2. r <- u mod v, u <- v, v <- r, loop to A1.
\ USES SCRATCH REGISTERS QQ RR (apart from result)

: VGCD          PP VVmove                \ <U> <V> -- <>
                SS VVmove
        BEGIN   PP @
        WHILE   SS PP VV/MOD      \ RR = remainder
                PP SS VVmove      \ u <- v
                RR PP VVmove      \ v <- (u mod v)_old
        REPEAT  ;


\ Calculate the largest number whose square is smaller than V, put it
\ in SS. We use Newton-Raphson : sqrt(b) = lim n-> An+1=(An+b/An)/2
\ USES SCRATCH REGISTERS QQ RR (apart from result)

: VSqrt        LOCAL V                         \ <V> -- <>
    V SS VVmove
    V Vcount-bits 1- 2/ 0 DO SS 1 V>> LOOP       \ Use l.s. half of the bits

    \ The starting value (in SS) starts above the root
    \  and decrease monotonically
    BEGIN
            V SS VV/MOD
            [ DEBUG ] [IF] CR SS .VNUM [THEN]
            SS QQ VV<= 0= WHILE                 \ Result in SS
            QQ SS VV+
            QQ 2 VS/MOD DROP
            QQ SS VVmove
    REPEAT
;

\ Bring the Vnumber V to the power n, result in SS
\ V =: V^n
: VS^                   \ V,n -- <>
    SS 1V!
    0 DO
         DUP SS VV*
         PP SS VVmove
    LOOP
    DROP
;


\ Calculate the largest number whose root is smaller than V, put it
\ in SS. We use Newton :
\ X <-- ( X*(n-1) +V/X^(n-1) )/n
\ Special case : sqrt X <-- ( X+V/X )/2
\ USES SCRATCH REGISTERS PP QQ RR (apart from result)

: VRoot                \ V,n --
    LOCAL n
    LOCAL VV
    VV SS VVmove

    \ Use slightly less than the n-th part of the bits to
    \  shift SS right.
    \ This assures that starting value (in SS) is close to
    \  but above the root
    VV Vcount-bits 1- n / n 1- * 0 ?DO SS 1 V>> LOOP

    \  Newton garantees that SS decrease monotonically
    [ DEBUG ] [IF] CR SS .VNUM [THEN]
    \  because the derivative doesnot change sign in the interval
    BEGIN
            SS PP VVmove                \ Use PP to remember, nobody uses it
            VV QQ VVmove
            n 1- 0 ?DO QQ SS VV/MOD LOOP
            SS n 1- VS*
            SS QQ VV+
            SS n VS/MOD DROP             \ SS is now iterated value
            [ DEBUG ] [IF] CR SS .VNUM [THEN]
            PP SS VV<= 0= WHILE         \ Still monotonic?
    REPEAT
    [ DEBUG ] [IF] CR SS .VNUM [THEN]
;


                              ( * End of Source * )
