        REVISION -jacobi " Number Theoretics   Version 1.00 "


\ The jacobi symbol of a single with a variable length number.

\ The following comment words only serve to show the amount
\ on the stack
\ All stack amount compare at two matching brackets
\ Example {1 1 {1 DROP }  1 2 }2
: {  ; IMMEDIATE : }  ; IMMEDIATE  \ Nothing on stack
: {1 ; IMMEDIATE : }1 ; IMMEDIATE  \ 1 item on stack etc.
: {2 ; IMMEDIATE : }2 ; IMMEDIATE
: {3 ; IMMEDIATE : }3 ; IMMEDIATE

( Remove all multiples of 4 from number u giving u' )
: KillFours    ( u -- u' )
     {1 BEGIN { DUP 3 AND 0= }1 WHILE   \ Kill multiples of 4 in u1'
            2 RSHIFT
        REPEAT
     }1
;

\ Like MOD but for unsigned numbers
\ Works also for both  > MAX-N
: UMOD >R U>D R> SM/REM DROP ;

\    ?Jacobi ( u1, u2 -- -1/0/+1 )
\    meaning  0 : u1 and u2 have a common factor
\            -1 : no common factor, u1 is no q.r. of u2
\            +1 :  "    "    "    , u1 i a q.r. of u2

\     Properties of Jacobi used in this calculation :
\     ( 0, m )   = 0
\     ( 1, m )   = 1
\     ( 2.a, 8*m+i ) = (a,8*m+i) for i=1,7
\     ( 2.a, 8*m+i ) = -(a,8*m+i) for i=3,5
\     ( x*m+n, m ) = (n, m) or ( n,m) =(n MOD m, m )
\     If both m and n odd :
\         ( n, m ) = ( m, n ) for m=4*x+1 or n=4*x+1
\         ( n, m ) = -( m, n ) for m=4*x+3 and n=4*x+3
\
\    Information : Knuth part 3 pg. 396.
\    ?Jacobi works on two single cell unsigned numbers.
\    It is assumed that u2 is odd.

: ?Jacobi-small  {2  ( u1, u2 -- -1/0/+1 )
     { DUP 1 AND 0= }1 IF ABORT THEN \
     {2 2DUP UMOD ROT DROP }2   \ Subtract multiples of u2 from u1
     {2                         \ Leave u2 u1'
     { DUP 0= }1 IF             \ Break recursion if u1' is 0
          {2 NIP }1 EXIT        \ This is the Jacobi result
        THEN
     }2                         \ u2 u1'
     {1 KillFours }1            \ u2 u1''
     { DUP 1 = }1 IF            \ Break recursion if u1' is 1
          {2 NIP }1 EXIT        \ This is the Jacobi result
     THEN
                                \ u2 u1''
     { DUP 1 AND }1 IF          \ u1'' odd
        { 2DUP AND 3 AND 3 = }1 >R  \ Remember both were 3 mod 4
        {2 RECURSE }1           \  with swapped numbers!
        {1 R> IF NEGATE THEN }1 \ Extra factor -1 if both 3 mod 4
     ELSE
     {2
        {2 2/ OVER }3
        {2 RECURSE  }1          \ Numbers not swapped!
        SWAP                    \ JACOBI, u2
        {1 7 AND DUP 3 = SWAP 5 = OR }1 IF
           NEGATE               \ Take (2,u2) into account
        THEN
     }1
     THEN
}1 ;

\ ?Jacobi returns the Jacobi symbol for the Vnumber V and unsigned u
: ?Jacobi  ( V,u -- -1/0/+1 )
  SWAP  SS VVmove       \ Copy to scratch register
  SS OVER VS/MOD        \ Leaves u, V mod u
  SWAP ?Jacobi-small
;

\ May be tested by "X test1 X test2", all numbers resulting from
\ the first test must be followed by 1 in the second test
\ X may not be even!
 : test1 DUP  0 DO  I I * OVER  UMOD          . LOOP DROP ;
 : test2 DUP  0 DO  I . I OVER ?Jacobi-small  . LOOP DROP ;
