\ FILENAME    : pbase.frt
\ CONTENT     : Handles the factor base
\ LANGUAGE    : tForth
\ PROJECT     : hcc primes
\ CATEGORY    : EXAMPLE
\ AUTHOR      : AH 911108
\ LAST CHANGE : MH 911121 stackbug in AddPrimes
\ LAST CHANGE : MH 911120 deleted two .VNUM in AddStarts
\ LAST CHANGE : AH 911119 PBselect , #primes is bases, not value
\ LAST CHANGE : 911110 AH Adding double starters
\ LAST CHANGE : 911110 AH Deletion of false starters
\ LAST CHANGE : 911110 AH DEPTH protection removed
\ LAST CHANGE : 911108 AH AddOnePower works apart from stopping
\ LAST CHANGE : 911108 AH created

        REVISION -pbase " Prime base          Version 1.00 "

        PRIVATES


VALUE PrimeFilled PRIVATE

  0 =: %prime%                        \ get prime (power) from table
  0 =: %from%                         \ my own from
  1 =: %to%                           \ my own to
 -1 =: %add%                          \ puts prime (power) and log into table
  2 =: %clear%                        \ initialisation command
  3 =: %addressof%                    \ killer may always be needed
  4 =: %sizeof%                       \ size command
 12 =: %2log%                         \ get 2log of prime (power) from table
 13 =: %tostart%                      \ set Start to table
 14 =: %fromstart%                    \ get Start from table
 15 =: %primemark%                    \ get last prime, start of powers
 16 =: %empty%                        \
 17 =: %print%                        \ print the entry

        \ Message posters

\ FROM also posts %prime%
\ CLEAR posts     %clear%
\ 'OF posts       %addressof%
\ +TO posts       %add%
\ SIZEOF posts    %sizeof%
: %prime          %prime%     %VAR !  ;
: %from           %from%      %VAR !  ;
: %to             %to%        %VAR !  ;
: %clear          %clear%     %VAR !  ;
: %addressof      %addressof% %VAR !  ;
: %sizeof         %sizeof%    %VAR !  ;
: %add            %add%       %VAR !  ;
: %2log           %2log%      %VAR !  ;
: %tostart        %tostart%   %VAR !  ;
: %fromstart      %fromstart% %VAR !  ;
: %primemark      %primemark% %VAR !  ;
: %empty          %empty%     %VAR !  ;
: %print          %print%     %VAR !  ;


\ Auxiliary definition for PBASE
: GET-PRIME     \ <index> <addrof a primebase> -- <0> | <prime>
                LOCAL PB
                DUP PB @ > OVER 0 <= OR IF
                   DROP 0 EXIT
                ENDIF
                3 * 1- CELLS PB + @
; PRIVATE

\ Auxiliary defition for PBASE
: PUT-PRIME     \ <prime> <LOG> <addrof a prime base> --
        LOCAL PB
        PB @ 3 * 2+ CELLS PB + LOCAL FREE-RECORD
        PB @ MaxPrime = IF
           TRUE ABORT" Too many entries in prime base "
        ENDIF
        FREE-RECORD CELL+ !               \ Store log
        FREE-RECORD !                     \ Store prime
        1 PB +!
; PRIVATE


\ Auxiliary definition for PBASE
: CLEAR-PB               \ <addres of a prime base> --
   MaxPrime 3 * 2+ CELLS ERASE
;

\ Structure of a PBASE
\ PBFilled indicator, PrimeMark marks the last real prime (then powers)
\ MaxPrime record with field : primes (or prime power)
\                            : 2logarithms (or incremental)
\                            : starting points.
\ There are <PBfilled> records used of MaxPrime available
: PBASE         CREATE                          \ <> PBASE #<name># \> <>
                   MaxPrime 3 * 2+ CELLS ALLOT
                DOES>                           \ <??> <pin> -- <?>
                   %VAR @  0 %VAR !
                   CASE
                      %clear%      OF CLEAR-PB               ENDOF
                      %prime%      OF GET-PRIME              ENDOF
                      %addressof%  OF  ( Leave address)      ENDOF
                      %add%        OF PUT-PRIME              ENDOF
                      %2log%       OF SWAP 3 *     CELLS + @ ENDOF
                      %tostart%    OF SWAP 3 * 1 + CELLS + ! ENDOF
                      %fromstart%  OF SWAP 3 * 1 + CELLS + @ ENDOF
                      %sizeof%     OF @                      ENDOF
                      %primemark%  OF CELL+ @                ENDOF
                      DUP DEC. TRUE ABORT" PBASE: unknown message"
                   ENDCASE ;


        PBASE PrimeBase



\ ?Criterion tests whether p deserves to be in the primebase
: ?Criterion    ( p -- flag )
  DUP ?Prime 0= IF              \ Sorry Sir, no dogs allowed
     DROP FALSE EXIT
  THEN
   TheNumber SWAP ?Jacobi        \ Is it a quadratic residu
  1 =
; PRIVATE

\ Make the first part of the factorbase : the primes
\ Auxiliary for CreatePrimeBase
: AddPrimes     ( -- )
        2 1 %add PrimeBase         \ All odd numbers are q.r. modulo 2
        2                          \ First candidate
        PrimeFilled 1 DO           \ Want another PrimeFilled-1
           BEGIN                   \ Increment until proper one
               1+ DUP
               ?Criterion 0= WHILE
           REPEAT
           DUP DUP 1- Scount-bits %add PrimeBase
       LOOP
       DROP
;

\ Add the data of the powers of a single prime to the factorbase
\ The index of that prime is <Index>
\ Auxiliary for CreatePrimeBase
: AddOnePower   ( Index-- )
    DUP %prime PrimeBase LOCAL P
        %2log PrimeBase LOCAL 2LOG
        %primemark PrimeBase %prime PrimeBase LOCAL MaxBase
    P
    BEGIN
        P * DUP                 \ Next power
        DUP MaxBase <= WHILE
        DUP 1- Scount-bits      \ The 2 logarithm
        2LOG OVER TO 2LOG       \ Remember it
        -                       \ Incremental effect of power
        %add PrimeBase
    REPEAT
    2DROP
; PRIVATE

\ Add the data of the powers of all primes already present to the factorbase
\ Also remember the mark between primes and powers
\ Auxiliary for CreatePrimeBase
: AddPowers             ( -- )
    %addressof PrimeBase LOCAL PB
    PB @ PB CELL+ !             \ Remember PrimeMark
    PB @ 1+ 1 DO
        I AddOnePower
    LOOP
;

\ I1 is an index in the sieve and SS the number represented by it.
\ Finds out whether the number SS is divisable by some prime(power),
\ for then is I1 a starter for this prime(power)
\ I2 is the part of the factorbase to be inspected for starts
\ Auxiliary for AddStarts
: MatchStarts   ( SS,I1,I2 -- )
    LOCAL #PrimaryPart
    LOCAL SieveIndex
    LOCAL SS
    0 LOCAL PrimePower
    #PrimaryPart 1+ 1 DO
        I %prime PrimeBase TO PrimePower
        I %fromstart PrimeBase 0= IF
            SS PrimePower VSMOD 0= IF
                SieveIndex I %tostart PrimeBase
            THEN
        ELSE                    \ There may be second starts
            SieveIndex PrimePower <= IF
                SS PrimePower VSMOD 0= IF
                    I %prime PrimeBase
                    I %2log  PrimeBase
                    %add PrimeBase          \ Another entry for this prime
                    SieveIndex %sizeof PrimeBase %tostart PrimeBase
                THEN
            THEN
        THEN
    LOOP
;

\ Add the start values for all entries of the PrimeBase
\ The method may seem peculiar at first.
\ We generate the numbers corresponding to the first MaxBase
\ entries of the sieve in sequence, and then look whether it happens to
\ be a starter. This is because generating takes the most time.
\ We know that the start is one of (1..MaxBase)
\ Auxiliary for CreatePrimeBase
: AddStarts                   ( -- )
    %sizeof PrimeBase LOCAL #Primary    \ Entries with primary start
    %primemark PrimeBase %prime PrimeBase LOCAL MaxBase

    TheRoot SS VVmove
    SS 1 VS+    SS SS VV*
    [ DEBUG ] [IF] CR PP .VNUM [THEN]
    [ DEBUG ] [IF] CR TheNumber .VNUM [THEN]
    PP TheNumber VV-
    PP SS VVmove            \ SS now corresponds to first element of sieve
    MaxBase 1+ 1 DO
        SS I #Primary MatchStarts
        SS TheRoot VV+
        SS TheRoot VV+
        SS I I + 1+ VS+     \ SS now corresponds to next element of sieve
    LOOP
;

\ Later we discovered that some of the powers actually are not starters
\ So here is an option to delete entries of the PrimeBase
: DeletePrimeBaseEntry  ( I -- )
    LOCAL Index
    %addressof PrimeBase LOCAL PB
    %sizeof PrimeBase LOCAL Len
    PB  2 Index 3 * + CELLS +   \ I+1 the entry : source
    DUP 3 CELLS -                   \ Destination
    Len Index - 3 * CELLS             \ Amount
    CMOVE                      \ Intelligent CMOVE required
    -1 PB +!
;


\ Later we discovered that some of the powers actually are not starters
\ So this check is an add on
\ Delete of all entries with no starter
\ Start at the end to prevent problems with changing length
\ If the Primes have no starter, it is a programming error
: CheckPrimeBase
    %sizeof PrimeBase LOCAL Index
    BEGIN
        Index 0 > WHILE
        Index %fromstart PrimeBase 0= IF
            Index %primemark PrimeBase <= ABORT" No starter for qr prime"
            Index DeletePrimeBaseEntry
        THEN
        -1 +TO Index
    REPEAT
;

\ Calculates the global variable PrimeFilled indicating the
\ number of primes actually occurring in the table
: CalculatePrimeFilled
    \ The following needs careful experimentation
    \ The PBSelect table has to be in ascending order
    TheNumber Vcount-bits LOCAL #BITS

    PBSelect
    BEGIN @+ #BITS < WHILE
          CELL+
    REPEAT
    @ TO PrimeFilled
; PRIVATE

\ Create a factor base into PrimeBase for TheNumber
\ The PrimeFilled is recalculated because TheNumber may have changed.
: CreatePrimeBase              ( -- )

    CalculatePrimeFilled
    %clear PrimeBase
    AddPrimes
    AddPowers
    AddStarts
    CheckPrimeBase
;

DEPRIVE
