
\ The infinite precision prime test for Vnumbers

\ AH 911108
\  LAST CHANGE : 931122 AH Use ZZ instead of SS
\  LAST CHANGE : 930302 AH SHIFT updates (changes to DPANS)
\  LAST CHANGE : ah 911117 Single precision prime test moved here
\  LAST CHANGE : ah 911108 cosmetic
\ Problems with the examples at the end did not originate here.
\ This version stands overnight testing of ?RABBIN

        REVISION -rabbin "--- Rabin oracle        Version 2    ---"

Vnumber ZZ      \ Local Scratch

\ Just an alias "most significant bit is up"
: MSB 0< ;

\ RR{new} = RR{old}^ (2^32) * VV^U modulo MM
\ Perform BIT/CELL (typical 32) bits of an exponentiation modulo MM
\ of a Vnumber VV to the bits of a single cell power U
\ RR is the intermediate product.
\ Loop invariant escapes me but it IS correct
\ (Something like RR*VV^(TOS>>I) modulo MM )
\ e.g. for V^S   starts with 1*VV^U, ends with RR*VV^0
\ Infinite loop whenever MM is not `COMPACT'ed
\ Result in RR ( remainder )
\ IF THIS IS DIFFICULT TRY TO UNDERSTAND IN COMBINATION WITH V^S
: V^bits     ( VV MM U -- )
   SWAP LOCAL MM
   SWAP LOCAL VV
   BITS/CELL 0 DO                \ For each bit
      RR RR VV*                 \ Square : Result in PP
      PP MM VV/MOD              \ Result in RR
      DUP MSB IF                \ If bit up
          RR VV VV*             \ Also multiply by VV : Result in PP
          PP MM VV/MOD          \ Result in RR
      ENDIF
      1 LSHIFT
  LOOP  DROP
;

\ RR = VV^U modulo MM
\ Put a Vnumber VV to a single cell power U modulo MM
\ Infinite loop whenever MM is not `COMPACT'ed
\ Result in RR ( remainder )
: V^S    ( VV MM U -- )
   RR 1V!
   V^bits
;

\ RR = VV^UU modulo MM
\ Put a Vnumber VV to the power UU modulo MM
\ Result in RR ( remainder )
\ Infinite loop whenever MM is not `COMPACT'ed
: V^V    ( VV UU MM -- )
   LOCAL MM
   LOCAL UU
   LOCAL VV
   RR 1V!
   1 UU @ DO
      VV  MM  I UU []CELL @    \ Prepare parameters
       V^bits
   -1 +LOOP

;

\ V{in} = V{out}*2^exponent
\ where V{out} modulo 2 is 0
\ Convert Vnum to yet another exponent, mantissa format
: EVEN-NORMALIZE      0 LOCAL exp                   \ <V> -- <exponent>
                DUP LOCAL V
                V COMPACT DUP @ 0= IF ABORT THEN          \  Protection
                1 CELLS +                      \ least significant cell
                BEGIN
                   DUP @ 1 AND 0= WHILE        \ l.s. bit 0
                   1 +TO exp
                   V 1 V>>
                REPEAT DROP
                V COMPACT
                exp
;


\ (?Rabbin) tells us whether TheNumber may be prime
\ FALSE : certainly not
\ TRUE : probably prime
Vnumber PROBE
: ?Rabbin  ( TheNumber -- flag )
  LOCAL TheNumber
  TheNumber ZZ VVmove
  ZZ 1 VS-             \ The random number put to this power
  ZZ EVEN-NORMALIZE LOCAL two's  \ Separate power of two
  1 PROBE !                \ Start PROBE up with a random number
  RANDOM PROBE CELL+ !
  PROBE ZZ TheNumber V^V   \ Result in RR

  RR @ 1 =  RR CELL+ @ 1 = AND IF       \ RR equal 1?
    TRUE EXIT                           \ Prime!
  ENDIF

  BEGIN
    RR 1 VS+
    RR TheNumber VV= IF
      TRUE EXIT                 \ Result -1 mod N -> probably prime
    THEN
  two's WHILE
    RR 1 VS-                    \ Restore RR
    RR RR VV*                   \ Square : Result in PP
    PP TheNumber VV/MOD         \ Result in RR
    -1 +TO two's
  REPEAT
  FALSE                         \ Not prime
;  PRIVATE

\ ?Prime tests whether the single precision number p is prime
\ Cases 0 1 are in fact illegal but return TRUE
: ?Prime        ( p -- flag )
  LOCAL P
  P 4 U< IF TRUE EXIT THEN       \ Prevent silly infinite loop
  P 1 AND 0= IF FALSE EXIT THEN  \ Handle even numbers ither than 2

  P 3 DO
    P I /MOD
    I < IF DROP TRUE LEAVE THEN
    0= IF FALSE LEAVE THEN
    2 +LOOP
;

\ ?prime tells us whether TheNumber may be prime
\ FALSE : certainly not
\ TRUE : almost surely prime
\ The uncertainty of 10^-x of ?Rabbin is raised to the
\ power #Tries
\ In practice we use it as a definitive test
: ?Vprime    ( V -- flag )
        LOCAL V

        V @ 0= IF FALSE EXIT THEN

        V @ 1 = IF              \ Single precision
           V CELL+ @ ?Prime     \ Simpler prime test
           EXIT
        ENDIF

        \ The following test only for large number above single precision
        TRUE                    \ Start with : may be prime
        #RabbinTries 0 DO
           V ?Rabbin AND         \ ALL rabbins must succeed..
           DUP 0= IF LEAVE THEN  \ if one fails we may stop
        LOOP
;

\ Not as good a test
\ Fermats small theorem, the resulting RR must be 1
\ for TheNumber to be prime.
: ?RAB2  ( TheNumber -- Vnum in RR )
  LOCAL TheNumber
  TheNumber ZZ VVmove
  ZZ 1 VS-             \ The random number put to this power
  1 PROBE !                \ Start PROBE up with a random number
  RANDOM PROBE CELL+ !
  PROBE ZZ TheNumber V^V   \ Result in RR
  RR COMPACT RR .VNUM
;

\ Use WT to select one of the following test primes
\ ?Rabbin should always give true for each of those
\ ?RAB should print 1
\ WT 1 = [IF]
\ TheNumber VREAD 5704689200685129054721
\ [THEN]
\
\ WT 2 = [IF]
\ TheNumber VREAD 37866809061660057264219253397
\ [THEN]
\
\ WT 3 = [IF]
\ TheNumber VREAD 59649589127497217
\ ZZ VREAD 5704689200685129054721
\ [THEN]
