\  LANGUAGE    : ANSI FORTH with tForth extensions
\  PROJECT     : Factoring
\  DESCRIPTION : the sieve presenting candidate solutions
\  CATEGORY    : EXAMPLE
\  AUTHOR      : Marcel Hendrix, 16:46:09  November 14, 1991
\  LAST CHANGE : 931127 AH use a protocol, total cleanup
\  LAST CHANGE : 920410 AH reinstalled WARNING for REGISTER
\  LAST CHANGE : 911119 AH Final test, long sieve installed, cosmetic
\  LAST CHANGE : 20:50:18  November 19, 1991, Marcel Hendrix, hardware links.
\  LAST CHANGE : 911117 ah cleanup van de indexen
\  LAST CHANGE : 21:25:18  November 16, 1991, Marcel Hendrix.
\  LAST CHANGE : 15:25:18  November 16, 1991, Marcel Hendrix.
\ /


        REVISION -sieve   "--- sieve               Version 3    ---"

        PRIVATES


                           -- MODULE 5 ** SIEVING --


\ Module 5; Sieving
\
\  We will use /sieve here, but this may be different on different machines.
\  The Sieve has length /sieve (bytes).
\
\  TheRoot is the root of TheNumber.
\
\  FindStarts has found for each prime(power) PP of Primebase a start S
\  such that (TheRoot+S)^2-TheNumber can be divided by PP. This could only be
\  done by trial and error. The results have been passed to module 3, for
\  everyone to use.
\
\ The task of the sieve is to add the log of this prime(power) to those
\ places, thus finding (by comparing with `CritFac') candidates that are
\ likely to be factored completely. The sieve is contained in `sieve', a
\ pointer to a dynamically allocated area with length `/sieve'. It represents
\ a data area starting at TheRoot+MyOffset.

\ The following data is duplicated for each sieve
 VALUE 'sieve   PRIVATE         CLEAR 'sieve
 VALUE /sieve   PRIVATE         CLEAR /sieve
 VALUE CritFac  PRIVATE         CLEAR CritFac
DVALUE MyOffset PRIVATE         CLEAR MyOffset
 VALUE SieverIn    \ Contains address of input channel
 VALUE SieverOut   \ Contains address of output channel

\ The protocol: > from main to sieve, < the other way
\ %>stop is accepted, but nobody is sending it (yet)
2 CONSTANT %>stop       \ >%
3 CONSTANT %>crit       \ >% >W
4 CONSTANT %>pbase      \ >% >M
5 CONSTANT %>offset     \ >% >D >W <W
6 CONSTANT %<work       \ <%
7 CONSTANT %<solution   \ <% <D

: CHANNEL-2@                                    \ <chan> --- <d>
        >S S CHANNEL-@ S> CHANNEL-@  SWAP ;

: CHANNEL-2!                                    \ <d> <chan> --- <>
        >S S CHANNEL-! S> CHANNEL-! ;


\  Allocate a (byte)sieve and fill it with zeroes.

\  Makes the sieve start at offset MyOffset
\  Has to communicate with main task to get parameters.

: ReceiveInstructions  \ PBase, Doffset, criterion  passed via channel
    TRUE LOCAL Stop?   \ Set to FALSE --> go!

\   ." Waiting in SieveReceive"
    SieverIn CHANNEL-@ CASE
\      CR ." Received " DUP . CR
       %>stop   OF STOP ENDOF
       %>crit   OF SieverIn CHANNEL-@  TO CritFac ENDOF
       %>pbase  OF
                   %addressof PrimeBase  MaxPrime 3 * 2+ CELLS
                   SieverIn CHANNEL-RECEIVE
                ENDOF
       %>offset OF
                   SieverIn CHANNEL-2@ TO MyOffset
                   SieverIn CHANNEL-@
                   \ Free only the first time
                   'sieve IF 'sieve FREE ?ALLOCATE THEN
                   AVAILABLE #10000 - MIN  \ don't choke on it
                   TO /sieve
\                  #1000 TO /sieve
                   /sieve ALLOCATE ?ALLOCATE  TO 'sieve
                   /sieve SieverOut CHANNEL-!
                   FALSE TO Stop?
                ENDOF
\      DUP DEC. ." ReceiveSolution: unknown message" STOP
    ENDCASE
\   {{ CR ." SIEVE" .S }}
    Stop?       \ Leave flag
;


\  SendSolution sends a solution to main.
\  This message is initiated from this side.

: SendSolution                                  \ <d> --- <>
\   {{ 2DUP ." Sending solution :" D. ." :" }}
    %<solution SieverOut CHANNEL-!
    SieverOut CHANNEL-2!
;


\ The Sieve represents the numbers (TheRoot+Offset+i)^2 - TheNumber
\ for i = 1 .. /sieve
\
\   - get Pin's prime(power) PP
\   - get Pin's start S from the PrimeBase
\   - get Pin's 2log T from the Primebase
\   - find first Index such that
\     Offset+Index - S can be divided by PP  (*)
\   - from Index to /sieve with step PP
\     add T to the Sieve element.
\
\  (*) To elaborate:
\
\       Offset + Index - S = n*PP  \
\                                   > Index = n*PP - (Offset-S)
\       0 <= Index < /sieve        /
\
\       It follows that Offset-S = a*PP+b with 0 <= b < PP,
\       so n must be:  a   if  b=0      ==> Index = 0
\                      a+1 if  b<>0     ==> Index = PP-b
\
\       Furthermore, we see that PP may not be larger than /sieve.
\
\       Concluding:
\         Index <- (Offset-S) mod PP
\         Index <> zero ==> Index <- (PP - Index)
\
\
\
\ Input :
\ - a double D the offset corresponding to first element in sieve
\ - the start for if D were 0
\ - the step length p
\ Output :
\ - the first index in this instance of the sieve that holds a number
\    divisible by p

:  CalcSieveStart  ( D, I, p  -- I )
        LOCAL pp
        LOCAL fromstart
           pp UM/MOD DROP               \ The skew
           \ The one is because in the sieve counting from zero, not 1
           fromstart 1- SWAP -          \ Subtract the skew
           DUP 0< IF pp + THEN          \ But the result could be outside
                                        \ the sieve
;


: SieveOnce ( dOffset Pin -- )
        0 LOCAL pp
        0 LOCAL t
        >S
         S  %prime     PrimeBase  TO pp         \ Prime(Power)
         S  %2log      PrimeBase  TO t
         ( dOffset) S> %fromstart PrimeBase  pp \ Parameters used by ..
         CalcSieveStart
        'sieve OVER + SWAP                      \ <first address> <first index>
        /sieve SWAP - pp /                      \ #cycles
        #40 /MOD SWAP >S                        \ unroll in sets of 40
           0 ?DO                                \ adjust Sieve at Index+k*PP
                 DUP t SWAP C+!   pp +          \ 40 times to flatten overhead
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +

( 10 )           DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +

( 20 )           DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +

( 30 )           DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
                 DUP t SWAP C+!   pp +
            LOOP
        S> 0 ?DO                                \ do the remainder entries
                 DUP t SWAP C+!   pp +
            LOOP
        DROP ; PRIVATE


\  do SieveOnce for each entry in Primebase.

: SieveAll ( dOffset -- )
\       7 EMIT
        %sizeof PrimeBase 1+ 1 ?DO
                                   2DUP I SieveOnce
                              LOOP
        2DROP
;


\ DOC Fish
\ (*
\  Fish for entries in the Sieve larger than FacCrit.
\  The approach below is too simple (slow), but is illustrative.
\
\  : Fish
\         LOCAL criterion                         \ <+n> --- <>
\         'sieve /sieve +  'sieve
\          ?DO
\              I C@ criterion > IF I 'sieve - 1+  MyOffset D+  SendAway
\                            ENDIF
\         LOOP ;
\ *)
\ ENDDOC


BITS/CELL 8 / =: BYTES/CELL     -- Won't work on some machines, adjust critmask

\ Create a 32-bit mask such that if ANDed with a 4-tuple of sieve entries,
\ at least one bit must be up in order for one of them to pass the criterion.
\ e.g. in binary   criterion : 0010 1011
\                  critmask  : 1110 0000 1110 0000 1110 0000 1110 0000
: critmask                                      \ <byte criterion> --- <w>
        DUP 0= IF 1- EXIT ENDIF
        Scount-bits  1- 2^x 1-                  \ test if possibly GTE
        DUP   8 LSHIFT OR
[ BYTES/CELL 2 > ] [IF] DUP #16 LSHIFT OR
                 [THEN]
        INVERT ; PRIVATE                        \ multiply the byte mask


: GOT-ONE?                                      \ <addr> <faccrit> --- <>
        OVER C@ <= IF                           \ [sieve] >= faccrit ?
                      'sieve -  1+ U>D MyOffset D+
                      SendSolution
                 ELSE DROP
                ENDIF ; PRIVATE

: GOT-FOUR?                                     \ <addr> <faccrit> --- <>
        >S
        BYTES/CELL
         0 DO
              DUP S GOT-ONE?
              1+
         LOOP
        DROP -S ; PRIVATE


: Fish                                          \ <+n> --- <>
        DUP      LOCAL criterion
        critmask LOCAL mask
        'sieve
        /sieve BYTES/CELL /MOD SWAP >S
           0 ?DO
                 DUP CELL+ SWAP
                 @  mask AND IF DUP CELL-  criterion  GOT-FOUR?
                          ENDIF
            LOOP
        S> 0 ?DO
                 DUP 1+ SWAP
                 C@ mask AND IF DUP 1-  criterion  GOT-ONE?
                          ENDIF
            LOOP
        DROP ;



\ `GoSieve' is intended to be run as a separate process before
\ the factorization starts. It need not be restarted for a new
\ factorization.
\  Start with:    LO-PRIO  ' GoSieve RUN Siever
\ where `Siever' is a FORTH-PROCESS

: GoSieve
     BEGIN
        BEGIN ReceiveInstructions WHILE REPEAT
        'sieve /sieve ERASE
        MyOffset SieveAll
        CritFac Fish            \ communicate with main using SendSolution
        %<work SieverOut CHANNEL-!      \ Ask for more work
     AGAIN
;


        DEPRIVE


                             ( * End of File * )
